import torch
import torch.nn as nn
from torch.nn import CrossEntropyLoss, MSELoss

from transformers.utils import logging

from transformers import GPT2PreTrainedModel, GPT2Model
from transformers.modeling_outputs import SequenceClassifierOutputWithPast


logger = logging.get_logger(__name__)

class GPT2hlcForSequenceClassification(GPT2PreTrainedModel):
    _keys_to_ignore_on_load_missing = [r"h\.\d+\.attn\.masked_bias", r"lm_head\.weight"]

    def __init__(self, config):
        super().__init__(config)
        self.freeze_model = config.freeze_model
        self.num_labels = config.num_labels
        self.finetuning_task = config.finetuning_task
        self.transformer = GPT2Model(config)

        if self.finetuning_task=='age':
            self.transform = nn.Linear(config.n_embd, config.n_embd)

        if not self.freeze_model and not self.finetuning_task=='ope':
            self.ln_f = nn.LayerNorm(config.n_embd, eps=config.layer_norm_epsilon)
        
        self.score = nn.Linear(config.n_embd, self.num_labels, bias=False)

        self.init_weights()

        # Model parallel
        self.model_parallel = False
        self.device_map = None

    def forward(
        self,
        input_ids=None,
        user_ids=None,
        past_key_values=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size,)`, `optional`):
            Labels for computing the sequence classification/regression loss. Indices should be in :obj:`[0, ...,
            config.num_labels - 1]`. If :obj:`config.num_labels == 1` a regression loss is computed (Mean-Square loss),
            If :obj:`config.num_labels > 1` a classification loss is computed (Cross-Entropy).
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        transformer_outputs = self.transformer(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=True if self.freeze_model else output_hidden_states,
            return_dict=return_dict,
        )
        hidden_states = transformer_outputs.hidden_states[11] if self.freeze_model else transformer_outputs[0]
        
        if self.freeze_model or self.finetuning_task=='ope' or self.finetuning_task=='user':
            logits = self.score(hidden_states) 
        elif self.finetuning_task=='age':
            self.score(self.transform(self.ln_f(hidden_states)))
        else:
            logits = self.score(self.ln_f(hidden_states))

        if input_ids is not None:
            batch_size, sequence_length = input_ids.shape[:2]
        else:
            batch_size, sequence_length = inputs_embeds.shape[:2]

        assert (
            self.config.pad_token_id is not None or batch_size == 1
        ), "Cannot handle batch sizes > 1 if no padding token is defined."
        if self.config.pad_token_id is None:
            sequence_lengths = -1
        else:
            if input_ids is not None:
                sequence_lengths = torch.ne(input_ids, self.config.pad_token_id).sum(-1) - 1
            else:
                sequence_lengths = -1
                logger.warning(
                    f"{self.__class__.__name__} will not detect padding tokens in `inputs_embeds`. Results may be "
                    f"unexpected if using padding tokens in conjunction with `inputs_embeds.`"
                )

        pooled_logits = logits[range(batch_size), sequence_lengths]

        loss = None
        if labels is not None:
            if self.num_labels == 1:
                if(self.finetuning_task=='document'):
                    labels_v2 = labels.to(self.dtype).view(-1)
                    preds = pooled_logits.view(-1)
                    indices = labels_v2!=-100
                    labels_v2 = labels_v2[indices]
                    preds = preds[indices]
                else:
                    labels = labels.long()
                    preds = pooled_logits.view(-1)
                    labels_v2 = labels.to(self.dtype).view(-1, self.num_labels)
            else:
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(pooled_logits.view(-1, self.num_labels), labels.view(-1))

        if not return_dict:
            output = (pooled_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutputWithPast(
            loss=loss,
            logits=pooled_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )
